from __future__ import absolute_import
from typing import Dict, List

from aoa.api.iterator_base_api import IteratorBaseApi
from aoa.api.trained_model_artefacts_api import TrainedModelArtefactsApi

import uuid


class TrainedModelApi(IteratorBaseApi):

    path = "/api/trainedModels/"
    type = "TRAINED_MODEL"

    def _get_header_params(self):
        header_vars = ['AOA-Project-ID', 'Content-Type', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            'application/json',
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]

        return self.generate_params(header_vars, header_vals)

    def find_dataset(self, trained_model_id: str, projection: str = None):
        """
        returns dataset of a trained model

        Parameters:
           trained_model_id (str): trained model id(uuid)
           projection (str): projection type

        Returns:
            (dict): dataset
        """
        query_vars = ['projection']
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=self.path + trained_model_id + '/dataset',
            header_params=self._get_header_params(),
            query_params=query_params)

    def find_events(self, trained_model_id: str, projection: str = None):
        """
        returns trained model events

        Parameters:
           trained_model_id (str): trained model id(uuid)
           projection (str): projection type

        Returns:
            (dict): events of trained model
        """
        query_vars = ['projection']
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=self.path + trained_model_id + '/events',
            header_params=self._get_header_params(),
            query_params=query_params)

    def find_by_model_id(self, model_id: str, projection: str = None):
        """
        returns a trained models by model id

        Parameters:
           model_id (str): model id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): trained models
        """
        query_vars = ['modelId', 'projection']
        query_vals = [model_id, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=self.path + 'search/findByModelId',
            header_params=self._get_header_params(),
            query_params=query_params)

    def find_by_model_id_and_status(self, model_id: str, status: str, projection: str = None):
        """
        returns a trained models by model id

        Parameters:
           model_id (str): model id(uuid) to find
           status (str): model status
           projection (str): projection type

        Returns:
            (dict): trained models
        """
        query_vars = ['modelId', 'status', 'projection']
        query_vals = [model_id, status, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=self.path + 'search/findByModelIdAndStatus',
            header_params=self._get_header_params(),
            query_params=query_params)

    def evaluate(self, trained_model_id: str, evaluation_request: Dict[str, str]):
        """
        evaluate a model

        Parameters:
           trained_model_id (str): trained model id(uuid) to evaluate
           evaluation_request (dict): request to evaluate trained model

        Returns:
            (dict): job
        """
        self.required_params(['datasetId'], evaluation_request)

        return self.aoa_client.post_request(
            path=self.path + trained_model_id + '/evaluate',
            header_params=self._get_header_params(),
            query_params={},
            body=evaluation_request)

    def approve(self, trained_model_id: str, comments: str):
        """
        approve a trained model
        :param trained_model_id:  model version
        :param comments: approval comments
        :return:
        """
        approve_request = dict()
        approve_request['comments'] = comments

        return self.aoa_client.post_request(
            path=self.path + trained_model_id + '/approve',
            header_params=self._get_header_params(),
            query_params={},
            body=approve_request)

    def reject(self, trained_model_id: str, comments: str):
        """
        reject a trained model
        :param trained_model_id:  model version
        :param comments: approval comments
        :return:
        """
        reject_request = dict()
        reject_request['comments'] = comments

        return self.aoa_client.post_request(
            path=self.path + trained_model_id + '/reject',
            header_params=self._get_header_params(),
            query_params={},
            body=reject_request)

    def deploy(self, trained_model_id: str, deploy_request: dict):
        """
        deploy a trained model
        :param trained_model_id:  model version
        :param deploy_request: deployment request
        :return:
        """
        self.required_params(['type'], deploy_request)

        return self.aoa_client.post_request(
            path=self.path + trained_model_id + '/deploy',
            header_params=self._get_header_params(),
            query_params={},
            body=deploy_request)

    def retire(self, trained_model_id: str, retire_request: dict):
        """
        retire a trained model
        :param trained_model_id:  model version
        :param retire_request: retire request
        :return:
        """
        self.required_params(['deploymentId'], retire_request)

        return self.aoa_client.post_request(
            path=self.path + trained_model_id + '/retire',
            header_params=self._get_header_params(),
            query_params={},
            body=retire_request)



import json
import pandas as pd
from typing import Dict

from teradataml import get_connection, DataFrame

ct_query = """
CT {} (
    column_name VARCHAR(128), 
    stats JSON, 
    update_ts TIMESTAMP)
UNIQUE PRIMARY INDEX ( column_name );
"""

merge_query = """
MERGE {} target
     USING {} source
       ON target.column_name = source.column_name
     WHEN MATCHED THEN
       UPD SET stats = source.stats, update_ts = source.update_ts
     WHEN NOT MATCHED THEN
       INS (source.column_name, source.stats, source.update_ts);
"""
temp_table = "aoa_stats_temp"


def save_feature_stats(features_table, stats: Dict, **kwargs):
    conn = get_connection()
    conn.execute(
        f"CREATE VOLATILE TABLE {temp_table} AS {features_table} WITH NO DATA ON COMMIT PRESERVE ROWS;")
    conn.execute(f"INS {temp_table} (?,?,CURRENT_TIMESTAMP);", [
                 [f, json.dumps(stats[f])] for f in stats])
    conn.execute(merge_query.format(features_table, temp_table))
    conn.execute(f"DROP TABLE {temp_table};")


def get_feature_stats(features_table):
    fs = DataFrame.from_query(f"SEL * FROM {features_table}")
    fs = fs.to_pandas().reset_index()
    fs = fs.drop(fs.columns.difference(["column_name", "stats"]), axis=1)
    fs = fs.set_index("column_name")
    fs = pd.Series(fs.stats).to_dict()
    return {k: json.loads(fs[k]) for k in fs}


def create_features_stats_table(features_table):
    conn = get_connection()
    conn.execute(ct_query.format(features_table))


from __future__ import absolute_import
from typing import Dict

from aoa.api.iterator_base_api import IteratorBaseApi


class DatasetApi(IteratorBaseApi):
    path = "/api/datasets/"
    type = "DATASET"

    def _get_header_params(self):
        header_vars = ['AOA-Project-ID', 'Content-Type', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            'application/json',
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]

        return self.generate_params(header_vars, header_vals)

    def find_by_name(self, dataset_name: str, projection: str = None):
        """
        returns a dataset of a project by name

        Parameters:
           dataset_name (str): dataset name(string) to find
           projection (str): projection type

        Returns:
            (dict): datasets
        """
        query_vars = ['name', 'projection']
        query_vals = [dataset_name, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=self.path + "search/findByName",
            header_params=self._get_header_params(),
            query_params=query_params)

    def find_by_dataset_template_id(self, dataset_template_id: str, projection: str = None, page: int = None, size: int = None, sort: str = None):
        """
        returns a dataset of a project by name

        Parameters:
           dataset_template_id (str): dataset_template_id
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): datasets
        """
        query_vars = ['datasetTemplateId', 'projection', 'page', 'size', 'sort']
        query_vals = [dataset_template_id, projection, page, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            path=self.path + "search/findByDatasetTemplateId",
            header_params=self._get_header_params(),
            query_params=query_params)

    def render(self, id: str) -> Dict:
        """
        returns a rendered dataset

        Parameters:
           id (str): dataset id

        Returns:
            (dict): rendered dataset
        """

        return self.aoa_client.get_request(
            path=self.path + id + "/render",
            header_params=self._get_header_params(),
            query_params={})

    def save(self, dataset: Dict[str, str]):
        """
        register a dataset

        Parameters:
           dataset (dict): dataset to register

        Returns:
            (dict): dataset
        """
        self.required_params(['name', 'metadata'], dataset)

        return self.aoa_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=dataset)




from teradataml import create_context, get_connection
import os
import logging

logger = logging.getLogger(__name__)


def aoa_create_context(database: str = None, temp_database: str = None):
    """
    Creates a teradataml context if one does not already exist.
    Most users should not need to understand how we pass the environment variables etc for dataset connections. This
    provides a way to achieve that and also allow them to work within a notebook for example where a context is already
    present.

    We create the connection based on the following environment variables which are configured automatically by the
    aoa based on the dataset connection selected:

        AOA_CONN_HOST
        AOA_CONN_USERNAME
        AOA_CONN_PASSWORD
        AOA_CONN_LOG_MECH
        AOA_CONN_DATABASE

    :param database: default database override
    :param temp_database: temp database override
    :return: None
    """
    if get_connection() is None:
        if not database:
            database = os.getenv("AOA_CONN_DATABASE")

        create_context(host=os.environ["AOA_CONN_HOST"],
                       username=os.environ["AOA_CONN_USERNAME"],
                       password=os.environ["AOA_CONN_PASSWORD"],
                       logmech=os.getenv("AOA_CONN_LOG_MECH", "TDNEGO"),
                       database=database if database else None,
                       temp_database_name=temp_database if temp_database else None)
    else:
        logger.info("teradataml context already exists. Skipping create_context.")



from teradataml import DataFrame


def store_byom_tmp(context, table: str, model_version: str, model_bytes: bytes):
    """
    Store a byom model in a temporary table so we can use it

    :param context: get_context()
    :param table: temp table name to use
    :param model_version: model version
    :param model_bytes: model bytes
    :return: DataFrame of model
    """
    context.execute(f"""
        CREATE VOLATILE TABLE {table}(
            model_version VARCHAR(255),
            model_id VARCHAR(255),
            model BLOB(2097088000)
        ) ON COMMIT PRESERVE ROWS;
        """)

    context.execute(f"INSERT INTO {table}(model_version, model) values(?,?)",
                    (model_version, model_bytes))

    return DataFrame.from_query(f"""
        SELECT model_version as model_id, model FROM {table} 
            WHERE model_version='{model_version}'
        """)


def check_if_table_exists(context, database: str, name: str, table_kind: str):
    """
    Check if the table/view/etc exists in the database

    :param context: sqlalchemy context/engine
    :param database: the database
    :param name: the table/view/etc name
    :param table_kind: 'V' for view, 'T' for table etc
    :return: (bool) True if exists, False otherwise
    """
    rs = context.execute(f"""
    SELECT * FROM DBC.TABLES WHERE 
        TABLENAME ='{name}' AND 
        tablekind = '{table_kind}' AND 
        databasename='{database}'
    """)

    return rs.rowcount >= 1



from teradataml import DataFrame, get_connection
from typing import List

import json
import sys


def save_evaluation_metrics(partition_df: DataFrame, metrics: List):
    """
    :param partition_df: teradata dataframe containing at least ["partition_id", "partition_metadata", "num_rows"]
    :param metrics: list of metrics to normalize and report
    :return: None
    """
    total_rows = int(partition_df.select(["num_rows"]).sum().to_pandas().iloc[0])

    metrics_sql = [f"SUM(CAST(partition_metadata AS JSON).JSONExtractValue('$.metrics.{metric}') * num_rows/{total_rows}) AS {metric}" for metric in metrics]
    joined_metrics_sql = ','.join(metrics_sql)
    metrics = DataFrame.from_query(f"SELECT {joined_metrics_sql} FROM {partition_df._table_name}").to_pandas()

    metrics = {metric: "{:.2f}".format(metrics.iloc[0][metric]) for metric in metrics}

    with open("artifacts/output/metrics.json", 'w+') as f:
        json.dump(metrics, f, indent=2)


def save_metadata(partition_df: DataFrame):
    """
    create statistic summaries based on the provided dataframe produced via training or evaluation

    partitions.json is {
        "<partition1 key>": <partition1_metadata>,
        "<partition2 key>": <partition2_metadata>,
        ...
    }

    data_stats.json is {
        "num_rows": <num_rows>,
        "num_partitions": <num_partitions>
    }

    :param partition_df: teradata dataframe containing at least ["partition_id", "partition_metadata", "num_rows"]
    :return: None
    """

    total_rows = int(partition_df.select(["num_rows"]).sum().to_pandas().iloc[0])

    metadata_df = partition_df.select(["partition_id", "partition_metadata", "num_rows"]).to_pandas()

    metadata_dict = {r["partition_id"]: json.loads(r["partition_metadata"])
                     for r in metadata_df.to_dict(orient='records')}

    with open("artifacts/output/partitions.json", 'w+') as f:
        json.dump(metadata_dict, f, indent=2)

    data_metadata = {
        "num_rows": total_rows,
        "num_partitions": int(metadata_df.shape[0])
    }

    with open("artifacts/output/data_stats.json", 'w+') as f:
        json.dump(data_metadata, f, indent=2)


def cleanup_cli(model_version: str, models_table: str = "aoa_sto_models"):
    """
    cli uses model version of "cli" always. We need to cleanup models table between runs.
    A better solution would be for the cli to write to a different table completely and just "recreate" on each run

    :param model_version: the model version being executed
    :param models_table: the models table for cleanup (default is aoa_sto_models)
    :return: None
    """
    if model_version == "cli":
        get_connection().execute("DELETE FROM {table} WHERE model_version='cli'".format(table = models_table))


def check_sto_version():
    """
    Check Python version In-Vantage against the version where this function is running,
    if it's incompatible raise an exception

    :return: None
    """
    version_query = """
    SEL DISTINCT ver
    FROM SCRIPT(
        SCRIPT_COMMAND('python3 -c "import sys; print(\\".\\".join(map(str, sys.version_info[0:2])))"')
        RETURNS('ver VARCHAR(10)') 
    );
    """
    local_version = '.'.join(map(str, sys.version_info[0:2]))
    result = get_connection().execute(version_query)
    if result.rowcount != 1:
        raise Exception('Different STO configuration on different nodes, please contact your system administrator')
    remote_version = next(iter(result))[0]
    if local_version != remote_version:
        raise Exception('Python versions not matching, local: {local}, In-Vantage: {remote}'.format(local = local_version, remote = remote_version))


def collect_sto_versions(raise_diff_config_exception=True):
    """
    Collects Python and packages information from In-Vantage installation

    :param raise_diff_config_exception: whether raise an exception if different Python versions are detected on different AMPs of Vantage system
    :return: Dict with python_version and packages versions
    """
    python_version_query = """
    SEL DISTINCT ver
    FROM SCRIPT(
        SCRIPT_COMMAND('python3 -c "import sys; print(sys.version.replace(\\"\\n\\",\\" \\"))"')
        RETURNS('ver VARCHAR(100)') 
    );
    """
    result = get_connection().execute(python_version_query)
    if result.rowcount != 1 and raise_diff_config_exception:
        raise Exception('Different STO configuration on different nodes, please contact your system administrator')
    python_version = next(iter(result))[0]

    packages_version_query = """
    SEL DISTINCT pkg
    FROM SCRIPT(
        SCRIPT_COMMAND('python3 -c "import pkg_resources; [print(pkg) for pkg in pkg_resources.working_set]"')
        RETURNS('pkg VARCHAR(100)') 
    );
    """
    result = get_connection().execute(packages_version_query)
    packages = {}
    for row in result:
        pair = row[0].split(" ")
        packages[pair[0]] = pair[1]

    return {'python_version' : python_version, 'packages' : packages}


def get_joined_models_df(data_table: str,
                         model_artefacts_table: str,
                         model_version: str,
                         partition_id: str = "partition_id"):
    """
    Joins the dataset which is to be used for scoring/evaluation with the model artefacts and appends the model_artefact
    to the first row with the column name 'model_artefact'.

    Args:
        data_table: the table/view of the dataset to join
        model_artefacts_table: the model artefacts table where the model artefacts are stored
        model_version: the model version to use from the model artefacts
        partition_id: the dataset partition_id

    Returns:
        DataFrame
    """
    query = f"""
    SELECT d.*, CASE WHEN n_row=1 THEN m.model_artefact ELSE null END AS model_artefact 
        FROM (SELECT x.*, ROW_NUMBER() OVER (PARTITION BY x.{partition_id} ORDER BY x.{partition_id}) AS n_row FROM {data_table} x) AS d
        LEFT JOIN {model_artefacts_table} m
        ON d.{partition_id} = m.partition_id
        WHERE m.model_version = '{model_version}'
    """

    return DataFrame(query=query)



from typing import List, Dict
import os


class DatasetInfo(object):

    def __init__(self, sql: str = None,
                 entity_key: str = None,
                 feature_names: List = None,
                 target_names: List = None,
                 feature_metadata: Dict = None,
                 predictions: Dict = None,
                 legacy_data_conf: Dict = None,
                 **kwargs):
        self.sql = sql
        self.entity_key = entity_key
        self.feature_names = feature_names
        self.target_names = target_names
        self.legacy_data_conf = legacy_data_conf

        if feature_metadata:
            self.feature_metadata_database = feature_metadata.get("database")
            self.feature_metadata_table = feature_metadata.get("table")
            self.feature_metadata_monitoring_group = feature_metadata.get("monitoringGroup")

        if predictions:
            self.predictions_database = predictions.get("database")
            self.predictions_table = predictions.get("table")

    @classmethod
    def from_dict(cls, rendered_dataset: Dict):
        if "type" in rendered_dataset and rendered_dataset["type"] == "CatalogBody":
            return cls(sql=rendered_dataset.get("sql"),
                       entity_key=rendered_dataset.get("entityKey"),
                       feature_names=rendered_dataset.get("featureNames"),
                       feature_metadata=rendered_dataset.get("featureMetadata"),
                       predictions=rendered_dataset.get("predictions"),
                       target_names=rendered_dataset.get("targetNames"))
        else:
            # set dict and legacy
            return cls(feature_metadata=rendered_dataset.get("featureMetadata"),
                       legacy_data_conf=rendered_dataset)

    def get_feature_metadata_fqtn(self):
        return f"{self.feature_metadata_database}.{self.feature_metadata_table}"

    def get_predictions_metadata_fqtn(self):
        return f"{self.predictions_database}.{self.predictions_table}"

    def is_legacy(self):
        return self.legacy_data_conf is not None


class ModelContext(object):

    def __init__(self, hyperparams: Dict,
                 dataset_info: DatasetInfo,
                 artefact_output_path: str = None,
                 artefact_input_path: str = None,
                 **kwargs):

        self.hyperparams = hyperparams
        self.artefact_output_path = artefact_output_path
        self.artefact_input_path = artefact_input_path
        self.dataset_info = dataset_info

        valid_var_keys = {"project_id", "model_id", "model_version", "job_id", "model_table"}
        for key in kwargs:
            if key in valid_var_keys:
                setattr(self, key, kwargs.get(key))

    @property
    def artefact_output_path(self):
        return self.__artefact_output_path

    @artefact_output_path.setter
    def artefact_output_path(self, artefact_output_path):
        if artefact_output_path and not os.path.isdir(artefact_output_path):
            raise ValueError(f"artefact_output_path ({artefact_output_path}) does not exist")

        self.__artefact_output_path = artefact_output_path

    @property
    def artefact_input_path(self):
        return self.__artefact_input_path

    @artefact_input_path.setter
    def artefact_input_path(self, artefact_input_path):
        if artefact_input_path and not os.path.isdir(artefact_input_path):
            raise ValueError(f"artefact_input_path ({artefact_input_path}) does not exist")

        self.__artefact_input_path = artefact_input_path




from aoa.cli.base_model import BaseModel
from aoa.context.model_context import ModelContext, DatasetInfo
import shutil
import os
import json
import logging
import importlib
import sys


class TrainModel(BaseModel):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def train_model_local(self,
                          model_id: str = None,
                          base_path: str = None,
                          rendered_dataset: dict = None):
        base_path = self.repo_manager.model_catalog_path if base_path is None else os.path.join(base_path, '')
        model_definitions_path = base_path + 'model_definitions/'

        if not os.path.exists(model_definitions_path):
            raise ValueError("model directory {0} does not exist".format(model_definitions_path))

        model_artefacts_path = ".artefacts/{}/".format(model_id)
        model_artefacts_output_path = ".artefacts/{}/output/".format(model_id)
        if os.path.exists(model_artefacts_path):
            self.logger.debug("Cleaning local model artefact path: {}".format(model_artefacts_path))
            shutil.rmtree(model_artefacts_path)

        os.makedirs(model_artefacts_output_path)

        try:
            if os.path.exists("./artifacts"):
                os.remove("./artifacts")

            if os.path.exists("./models"):
                os.remove("./models")

            os.symlink(model_artefacts_path, "./artifacts", target_is_directory=True)
            os.symlink(model_artefacts_output_path, "./models", target_is_directory=True)

            model_dir = model_definitions_path + BaseModel.get_model_folder(model_definitions_path, model_id)

            with open(model_dir + "/model.json", 'r') as f:
                model_definition = json.load(f)

            with open(model_dir + "/config.json", 'r') as f:
                model_conf = json.load(f)

            self.logger.info("Loading and executing model code")

            cli_model_kargs = self._BaseModel__get_model_varargs(model_id)

            context = ModelContext(dataset_info=DatasetInfo.from_dict(rendered_dataset),
                                   hyperparams=model_conf["hyperParameters"],
                                   artefact_output_path="artifacts/output",
                                   **cli_model_kargs)

            engine = self._BaseModel__get_engine(model_definition, "training")
            if engine == "python":
                sys.path.append(model_dir)
                training = importlib.import_module(".training", package="model_modules")
                training.train(context=context,
                               data_conf=rendered_dataset,
                               model_conf=model_conf,
                               **cli_model_kargs)

            elif engine == "sql":
                self.__train_sql(context, model_dir, rendered_dataset, model_conf, **cli_model_kargs)

            elif engine == "R":
                self._BaseModel__run_r_model(model_id, base_path, rendered_dataset, "train")

            else:
                raise Exception("Unsupported engine: {}".format(engine))

            self.logger.info("Artefacts can be found in: {}".format(model_artefacts_output_path))
            self.__cleanup()
        except ModuleNotFoundError:
            model_dir = model_definitions_path + BaseModel.get_model_folder(model_definitions_path, model_id)
            self.__cleanup()
            self.logger.error("Missing required python module, try running following command first:")
            self.logger.error("pip install -r {}/model_modules/requirements.txt".format(model_dir))
            raise
        except:
            self.__cleanup()
            self.logger.exception("Exception running model code")
            raise

    def __cleanup(self):
        if os.path.exists("./artifacts"):
            os.remove("./artifacts")
        if os.path.exists("./models"):
            os.remove("./models")

    def __train_sql(self, context, model_dir, data_conf, model_conf, **kwargs):
        from aoa.util import aoa_create_context
        from teradataml import remove_context, get_connection

        self.logger.info("Starting training...")

        aoa_create_context()

        sql_file = model_dir + "/model_modules/training.sql"
        jinja_ctx = {
            "context": context,
            "data_conf": data_conf,
            "model_conf": model_conf,
            "model_table": kwargs.get("model_table"),
            "model_version": kwargs.get("model_version"),
            "model_id": kwargs.get("model_id")
        }

        self._BaseModel__execute_sql_script(get_connection(), sql_file, jinja_ctx)

        remove_context()

        self.logger.info("Finished training")

        self.logger.info("Saved trained model")

        remove_context()
